"""
Support for metric units.
"""

import contextlib
import numbers
import typing

from .. import aliased
from .. import exceptions
from .. import symbolic
from . import _conversions
from . import _defined
from . import _dimensions
from . import _exceptions
from . import _reference


T = typing.TypeVar('T')


class Unit(symbolic.Expression):
    """A symbolic expression representing a physical unit."""

    def __init__(self, terms: typing.List[symbolic.Term]) -> None:
        super().__init__(terms)
        self._dimensions = None
        self._decomposed = None
        self._dimensionless = None
        self._quantity = None
        self._norms = dict.fromkeys(_reference.SYSTEMS)

    def normalize(self, system: str, quantity: str=None):
        """Represent this unit in base units of `system`.

        This is an instance-method version of `~normalize`. See that function
        for full documentation.
        """
        return normalize(self, system, quantity)

    def __mul__(self, other):
        """Called for self * other."""
        if isinstance(other, numbers.Real):
            if other == 1:
                return self
            raise ValueError(f"Cannot compute {self!r} * {other}") from None
        if self is other or super().__eq__(other):
            return unit_factory(super().__pow__(2))
        return _apply_operator(symbolic.product, self, other)

    def __rmul__(self, other):
        """Called for other * self."""
        if isinstance(other, numbers.Real):
            if other == 1:
                return self
            raise ValueError(f"Cannot compute {other} * {self!r}") from None
        return _apply_operator(symbolic.product, other, self)

    def __truediv__(self, other):
        """Called for self / other."""
        if isinstance(other, numbers.Real):
            if other == 1:
                return self
            raise ValueError(f"Cannot compute {self!r} / {other}") from None
        if self is other or super().__eq__(other):
            return unit_factory(1)
        return _apply_operator(symbolic.ratio, self, other)

    def __rtruediv__(self, other):
        """Called for other / self."""
        if isinstance(other, numbers.Real):
            if other == 1:
                return _apply_operator(symbolic.ratio, '1', self)
            raise ValueError(f"Cannot compute {other} / {self!r}") from None
        return _apply_operator(symbolic.ratio, other, self)

    def __pow__(self, other):
        """Called for self ** other."""
        if not isinstance(other, numbers.Real):
            return NotImplemented
        return unit_factory([term ** other for term in self])

    def __rpow__(self, other):
        """Called for other ** self."""
        return NotImplemented

    def __rshift__(self, other):
        """Compute the numerical factor to convert this unit to `other`.

        Examples
        --------
        Create two unit objects representing a meter and a centimeter, and
        compute their relative magnitude::

        >>> m = Unit('m')
        >>> cm = Unit('cm')
        >>> cm >> m
        0.01
        >>> m >> cm
        100.0

        As with `~metric._defined.NamedUnit`, these results are equivalent to
        the statement that there are 100 centimeters in a meter. However, this
        class also supports more complex unit expressions, and can therefore
        compute more complex ratios::

        >>> Unit('g * m^2 / s^2') >> Unit('kg * m^2 / s^2')
        0.001
        >>> Unit('g * mm^2 / s^2') >> Unit('kg * m^2 / s^2')
        1e-09
        >>> Unit('g * m^2 / day^2') >> Unit('kg * m^2 / s^2')
        1.3395919067215364e-13
        >>> Unit('g * au^2 / day^2') >> Unit('kg * m^2 / s^2')
        2997942777.7207007
        """
        if not isinstance(other, (str, Unit)):
            return NotImplemented
        return float(_conversions.conversion_factory(str(self), str(other)))

    def __rrshift__(self, other):
        """Reflected version of `__rshift__`."""
        if isinstance(other, str):
            return float(_conversions.conversion_factory(other, str(self)))
        return NotImplemented

    def __lshift__(self, other):
        """Compute the numerical factor to convert `other` to this unit."""
        if not isinstance(other, (str, Unit)):
            return NotImplemented
        return float(_conversions.conversion_factory(str(other), str(self)))

    def __rlshift__(self, other):
        """Reflected version of `__lshift__`."""
        if isinstance(other, str):
            return float(_conversions.conversion_factory(str(self), other))
        return NotImplemented

    def __eq__(self, other) -> bool:
        """Called for self == other.

        Two unit expressions are equal if they satisfy one of the following
        conditions:

        1. are identical
        1. have symbolically equal strings
        1. have symbolically equivalent strings

        For example, all of the following are true:

        * `Unit('N') == Unit('N')` (condition 1)
        * `Unit('m / s') == 'm / s'` (condition 2)
        * `'m / s' == Unit('m / s')` (condition 2)
        * `Unit('m / s') == 'm s^-1'` (condition 3)
        * `'m / s' == Unit('m s^-1)'` (condition 3)

        See Also
        --------
        `__and__`:
            Test equivalence of two unit expressions.
        `__or__`:
            Test consistency of two unit expressions.
        """
        if other is self:
            # If they are identical, they are equal.
            return True
        # Otherwise, they are equal iff their symbolic expressions are equal.
        return super().__eq__(other)

    def __and__(self, other):
        """Called for self & other.

        This method tests whether the given unit is equivalent to this unit. Two
        unit expressions are are equivalent if they satisfy one of the following
        conditions:

        1. are equal (see docstring at `__eq__`)
        1. differ only by dimensionless terms
        1. have a ratio of unity

        For example, all of the following are true:

        * `Unit('1 / s') & '# / s'` (condition 2)
        * `'1 / s' & Unit('# / s')` (condition 2)
        * `Unit('N') & 'kg * m / s^2'` (condition 3)
        * `'N' & Unit('kg * m / s^2')` (condition 3)

        Equivalence thus defines a notion of equality that does not require two
        unit expressions that represent equal physical quantities to have equal
        lexical representations.
        """
        if self == other:
            return True
        unity = (
            str(term) in _reference.UNITY
            for term in self.difference(other, symmetric=True)
        )
        if all(unity):
            # If the only terms that differ between the units are dimensionless
            # terms, we can declare them equal by inspection (i.e., without
            # computing their conversion factor).
            return True
        that = unit_factory(other)
        if self.dimensionless != that.dimensionless:
            # If one, and only one, of the units is dimensionless, they can't
            # possibly be equivalent.
            return False
        if set(self.decomposed) == set(that.decomposed):
            # If their base units are equal, the units are equivalent.
            return True
        if symbolic.equality(self.decomposed, that.decomposed):
            # If their base units produce equal expressions, the units are
            # equivalent.
            return True
        with contextlib.suppress(_exceptions.UnitConversionError):
            # If their numerical conversion factor is unity, they are
            # equivalent.
            return self << that == 1.0
        return False

    __rand__ = __and__
    """Called for other & self.

    Notes
    -----
    This is the reflected version of `~metric.Unit.__and__`. It exists to
    support equivalence comparisons between instances of `~metric.Unit` and
    objects of other types for which that comparison is meaningful, in cases
    where the other object is the left operand. The semantics are identical.
    """

    def __or__(self, other):
        """Called for self | other.

        This method tests whether the given unit is metrically consistent with
        this unit. Two units are metrically consistent if they satisfy one of
        the following conditions:

        1. are equal (see docstring at `__eq__`)
        1. are equivalent (see docstring at `__and__`)
        1. have the same dimension in at least one metric system
        1. have a defined conversion factor

        For example, all of the following are true:

        * `Unit('J') | 'erg'` (condition 3)
        * `'J' | Unit('erg')` (condition 3)
        * `Unit('J') | 'eV'` (condition 4)
        * `'J' | Unit('eV')` (condition 4)

        Consistency thus provides a way to determine if this unit can convert to
        the given unit without necessarily attempting the conversion.
        """
        if self == other or self & other:
            return True
        that = unit_factory(other)
        for system in _reference.SYSTEMS:
            defined = self.dimensions[system]
            given = that.dimensions.values()
            if defined and any(d == defined for d in given):
                return True
        if _equivalent_quantities(self, that):
            return True
        try:
            self << that
        except _exceptions.UnitConversionError:
            return False
        else:
            return True

    __ror__ = __or__
    """Called for other | self.

    Notes
    -----
    This is the reflected version of `~metric.Unit.__or__`. It exists to support
    consistency comparisons between instances of `~metric.Unit` and objects of
    other types for which that comparison is meaningful, in cases where the
    other object is the left operand. The semantics are identical.
    """

    @property
    def quantity(self):
        """A symbolic expression of this unit's metric quantity."""
        if self._quantity is None:
            self._quantity = _defined.get_quantity(self)
        return self._quantity

    @property
    def dimensionless(self):
        """True if this unit's dimension is '1' in all metric systems.
        
        Notes
        -----
        This property exists as a convenient shortcut for comparing this unit's
        dimension in each metric system to '1'. If you want to check whether
        this unit is dimensionless in a specific metric system, simply check the
        `dimensions` property for that system.
        """
        if self._dimensionless is None:
            self._dimensionless = all(
                dimension == '1'
                for dimension in self.dimensions.values()
            )
        return self._dimensionless

    @property
    def dimensions(self):
        """The physical dimension of this unit in each metric system."""
        if self._dimensions is None:
            guard = exceptions.Wrapper(self._dimensions_in)
            guard.catch(ValueError)
            mapping = {s: guard.compute(s) for s in _reference.SYSTEMS}
            self._dimensions = _dimensions.Dimensions(**mapping)
        return self._dimensions

    def _dimensions_in(self, system: typing.Literal['mks', 'cgs']):
        """Internal helper for `dimensions` property."""
        expression = symbolic.expression('1')
        systems = set()
        for term in self:
            named = _defined.NamedUnit(term.base)
            allowed = named.systems['allowed']
            dimension = (
                named.dimensions[system] if len(allowed) > 1
                else named.dimensions[allowed[0]]
            )
            expression *= symbolic.expression(dimension) ** term.exponent
            systems.update(allowed)
        if system in systems:
            return _dimensions.Dimension(expression)
        raise ValueError(
            f"Can't define dimension of {self!r} in {system!r}"
        ) from None

    @property
    def decomposed(self):
        """This unit's decomposition into base units, where possible."""
        if self._decomposed is None:
            self._decomposed = _decompose_unit(self)
        return self._decomposed


def _apply_operator(f, a, b):
    """Compute `f(a, b)` where at least `a` or `b` is a `~Unit`.

    This method will attempt to reduce each operand into base units before
    computing the result, in order to reduce the result as much as possible
    """
    if unitlike(a) and unitlike(b):
        return unit_factory(f(_decompose_unit(a), _decompose_unit(b)))
    if isinstance(a, Unit) and isinstance(b, numbers.Real):
        return unit_factory(f(a.decomposed, b))
    if isinstance(a, numbers.Real) and isinstance(b, Unit):
        return unit_factory(f(a, b.decomposed))
    return NotImplemented


def _decompose_unit(unit: typing.Union[str, Unit]):
    """Decompose the given unit into base units where possible."""
    decomposed = [
        part
        for term in symbolic.expression(unit)
        for part in _decompose_term(term)
    ]
    return symbolic.reduce(decomposed)


def _decompose_term(term: symbolic.Term):
    """Decompose a symbolic term into named units, if possible."""
    try:
        # Possible outcomes
        # - success: returns list of terms
        # - decomposition failed: returns `None`
        # - parsing failed: raises `UnitParsingError`
        # - metric system is ambiguous: raises `SystemAmbiguityError`
        current = _defined.NamedUnit(term.base).decomposed
    except (_exceptions.UnitParsingError, _exceptions.SystemAmbiguityError):
        # This effectively reduces the three failure modes listed above into
        # one result.
        current = None
    if current is None:
        # If the attempt to decompose this unit term failed or raised an
        # exception, our only option is to append the existing term to the
        # running list.
        return [term]
    # If the attempt to decompose this unit term succeeded, we need to
    # distribute the term's exponent over the new terms and append each to
    # the running list.
    return [base**term.exponent for base in current]


_units = aliased.MutableMapping()
"""Internal collection of singleton `~Unit` instances."""

def unit_factory(arg: typing.Union[str, Unit]) -> Unit:
    """Create a metric unit representing the given expression."""
    if isinstance(arg, Unit):
        return arg
    # Attempt to extract a string representing a single unit.
    if isinstance(arg, str):
        string = arg
    else:
        try:
            string = str(arg[0]) if len(arg) == 1 else None
        except TypeError:
            string = None
    if available := _units.get(string):
        # If the argument maps to an existing unit, return that unit.
        return available
    # First time through: create a symbolic expression from `arg`.
    expression = symbolic.expression(arg)
    # The canonical string representation of this unit is the expression
    # string of its unit symbols. For example, both 'm / s' and 'meter /
    # second' should map to 'm / s'.
    symbols = [_as_symbol(term) for term in expression]
    instance = Unit(symbols)
    name = str(instance)
    if name in _units:
        # It turns out that the argument corresponds to an existing unit.
        existing = _units[name]
        if isinstance(arg, str):
            # If the argument is a string, register the argument as an alias
            # for that unit so we can just retrieve it next time.
            _units.alias(name, arg)
        return existing
    # Create the initial mapping aliases for this unit.
    try:
        # If `name` corresponds to a named unit, register both the name and
        # symbol (e.g., 'centimeter' and 'cm').
        this = _defined.NamedUnit(name)
        key = (this.name, this.symbol)
    except _exceptions.UnitParsingError:
        # If attempting to parse a named unit from `name` failed, register
        # the canonical string and, if applicable, the string argument.
        key = (name, string) if string else name
    # Store and return the new instance.
    _units[key] = instance
    return instance


def normalize(
    unit: typing.Union[str, Unit],
    system: str,
    quantity: str=None,
) -> Unit:
    """Represent a unit in base units of `system`.
    
    Parameters
    ----------
    unit : string or `~Unit`
        The original unit. If `unit` is a string, this function will first
        convert it to an instance of `~Unit`.

    system : string
        The metric system in which to express the result.

    quantity : string, optional
        The physical quantity associated with `unit`. If `quantity` is omitted,
        this function will compute a suitable quantity from the base units of
        `unit`. Providing an explicit quantity may resolve ambiguity in certain
        cases (e.g., between 'velocity' and 'conductance' in CGS).

    Returns
    -------
    `~Unit`
    """
    this = unit_factory(unit)
    terms = (
        _defined.get_quantity(this) if quantity is None
        else symbolic.expression(quantity)
    )
    expression = [
        symbolic.power(_defined.UNITS[term.base][system], term.exponent)
        for term in terms
    ]
    return unit_factory(expression)


def _as_symbol(unit: symbolic.Term):
    """Represent `unit` in terms of its symbol."""
    this = _reference.NAMED_UNITS[unit.base]
    prefix = this['prefix']['symbol']
    base = this['base']['symbol']
    return symbolic.term(prefix + base) ** unit.exponent


def _equivalent_quantities(u0: Unit, u1: Unit):
    """True if two units have equivalent physical quantities."""
    try:
        q0 = _defined.get_quantity(u0)
        q1 = _defined.get_quantity(u1)
    except Exception:
        # NOTE: This is the same pattern that `numpy.array_equal` uses.
        # Basically, if anything goes wrong while trying to compute the physical
        # quantity of either unit, we give up and declare them not equivalent.
        return False
    return q0 == q1


UnitType = typing.TypeVar('UnitType', Unit, symbolic.Expressable)
"""Type variable for unit-like objects.

A unit-like object is an object that can represent a metric unit, if it
corresponds to a known metric unit. Formally, it is an object that can
instantiate the `~metric.Unit` class. This definition makes no guarantee that
the resulting instance will be valid.

See Also
--------
`~metric.UnitLike`
    The equivalent type alias.
`~metric.unitlike`
    The corresponding instance test.
"""


UnitLike = typing.Union[Unit, symbolic.Expressable]
"""Type alias for unit-like objects.

A unit-like object is an object that can represent a metric unit, if it
corresponds to a known metric unit. Formally, it is an object that can
instantiate the `~metric.Unit` class. This definition makes no guarantee that
the resulting instance will be valid.

See Also
--------
`~metric.UnitType`
    The equivalent type variable.
`~metric.unitlike`
    The corresponding instance test.
"""


def unitlike(this):
    """True if `this` can serve as a metric unit.
    
    Notes
    -----
    This function will return `True` if `this` is

    * an instance of `~metric.Unit` or `str`, or one of their subclasses
    * a strict instance of `~symbolic.Expression`

    The justification for the restriction in the latter case is to avoid a false
    positive when `this` is a `~metric` subclass of `~symbolic.Expression`
    (e.g., `~metric.Dimension`).

    See Also
    --------
    `~metric.UnitType`
        The type variable for unit-like objects.
    `~metric.UnitLike`
        The type alias for unit-like objects.
    """
    return (
        isinstance(this, (Unit, str))
        or type(this) is symbolic.Expression
    )


def reduction(unit: symbolic.Expressable, system: str=None):
    """Reduce the given unit expression, if possible.
    
    Notes
    -----
    This function is still experimental.
    """
    expression = symbolic.expression(unit)
    decomposed = []
    for term in expression:
        try:
            # NOTE: `_defined.NamedUnit.reduce` can return `None`
            current = _defined.NamedUnit(term.base).reduce(system=system)
        except (_exceptions.UnitParsingError, _exceptions.SystemAmbiguityError):
            current = None
        if current:
            decomposed.append(current**term.exponent)
    if not decomposed:
        return
    result = decomposed[0]
    for other in decomposed[1:]:
        result *= other.scale
    return result





"""
Support for defining and working with measurable objects.
"""

import numbers
import typing

import numpy
import numpy.typing

from .. import metric


@typing.runtime_checkable
class Type(metric.Type, typing.Protocol):
    """Structural protocol class for metric objects with data."""

    data: typing.Union[
        typing.Sequence[numbers.Real],
        numpy.typing.NDArray[numpy.integer],
        numpy.typing.NDArray[numpy.floating],
    ]





"""
All independently defined variables for the `metric` namespace.
"""

import typing

import numpy

from .. import aliased
from .. import container


T = typing.TypeVar('T')


_PREFIXES = [
    {'symbol': 'Y', 'name': 'yotta', 'factor': 1e+24},
    {'symbol': 'Z', 'name': 'zetta', 'factor': 1e+21},
    {'symbol': 'E', 'name': 'exa', 'factor': 1e+18},
    {'symbol': 'P', 'name': 'peta', 'factor': 1e+15},
    {'symbol': 'T', 'name': 'tera', 'factor': 1e+12},
    {'symbol': 'G', 'name': 'giga', 'factor': 1e+9},
    {'symbol': 'M', 'name': 'mega', 'factor': 1e+6},
    {'symbol': 'k', 'name': 'kilo', 'factor': 1e+3},
    {'symbol': 'h', 'name': 'hecto', 'factor': 1e+2},
    {'symbol': 'da', 'name': 'deca', 'factor': 1e+1},
    {'symbol': '', 'name': '', 'factor': 1e0},
    {'symbol': ' ', 'name': None, 'factor': 0.0},
    {'symbol': 'd', 'name': 'deci', 'factor': 1e-1},
    {'symbol': 'c', 'name': 'centi', 'factor': 1e-2},
    {'symbol': 'm', 'name': 'milli', 'factor': 1e-3},
    {'symbol': 'μ', 'name': 'micro', 'factor': 1e-6},
    {'symbol': 'n', 'name': 'nano', 'factor': 1e-9},
    {'symbol': 'p', 'name': 'pico', 'factor': 1e-12},
    {'symbol': 'f', 'name': 'femto', 'factor': 1e-15},
    {'symbol': 'a', 'name': 'atto', 'factor': 1e-18},
    {'symbol': 'z', 'name': 'zepto', 'factor': 1e-21},
    {'symbol': 'y', 'name': 'yocto', 'factor': 1e-24},
]


UNITY = {'#', '1'}
"""Strings that represent dimensionless units."""


SYSTEMS = {'mks', 'cgs'}
"""The metric systems known to this module."""


_UNITS = [
    {
        'symbol': 'm',
        'name': 'meter',
        'quantity': 'length',
    },
    {
        'symbol': 'au',
        'name': 'astronomical unit',
        'quantity': 'length',
    },
    {
        'symbol': 'Rs',
        'name': 'solar radius',
        'quantity': 'length',
    },
    {
        'symbol': 'g',
        'name': 'gram',
        'quantity': 'mass',
    },
    {
        'symbol': 'nuc',
        'name': 'nucleon',
        'quantity': 'mass number',
    },
    {
        'symbol': 'amu',
        'name': 'atomic mass unit',
        'quantity': 'mass number',
    },
    {
        'symbol': 's',
        'name': 'second',
        'quantity': 'time',
    },
    {
        'symbol': 'min',
        'name': 'minute',
        'quantity': 'time',
    },
    {
        'symbol': 'h',
        'name': 'hour',
        'quantity': 'time',
    },
    {
        'symbol': 'd',
        'name': 'day',
        'quantity': 'time',
    },
    {
        'symbol': 'A',
        'name': 'ampere',
        'quantity': 'current',
    },
    {
        'symbol': 'K',
        'name': 'kelvin',
        'quantity': 'temperature',
    },
    {
        'symbol': 'mol',
        'name': 'mole',
        'quantity': 'amount',
    },
    {
        'symbol': '#',
        'name': 'count',
        'quantity': 'number',
    },
    {
        'symbol': 'cd',
        'name': 'candela',
        'quantity': 'luminous intensity',
    },
    {
        'symbol': 'rad',
        'name': 'radian',
        'quantity': 'plane angle',
    },
    {
        'symbol': 'deg',
        'name': 'degree',
        'quantity': 'plane angle',
    },
    {
        'symbol': 'sr',
        'name': 'steradian',
        'quantity': 'solid angle',
    },
    {
        'symbol': 'Hz',
        'name': 'hertz',
        'quantity': 'frequency',
    },
    {
        'symbol': 'J',
        'name': 'joule',
        'quantity': 'energy',
    },
    {
        'symbol': 'erg',
        'name': 'erg',
        'quantity': 'energy',
    },
    {
        'symbol': 'eV',
        'name': 'electronvolt',
        'quantity': 'energy',
    },
    {
        'symbol': 'N',
        'name': 'newton',
        'quantity': 'force',
    },
    {
        'symbol': 'dyn',
        'name': 'dyne',
        'quantity': 'force',
    },
    {
        'symbol': 'Pa',
        'name': 'pascal',
        'quantity': 'pressure',
    },
    {
        'symbol': 'W',
        'name': 'watt',
        'quantity': 'power',
    },
    {
        'symbol': 'C',
        'name': 'coulomb',
        'quantity': 'charge',
    },
    {
        'symbol': 'statC',
        'name': 'statcoulomb',
        'quantity': 'charge',
    },
    {
        'symbol': 'statA',
        'name': 'statampere',
        'quantity': 'current',
    },
    {
        'symbol': 'statV',
        'name': 'statvolt',
        'quantity': 'potential',
    },
    {
        'symbol': 'e',
        'name': 'fundamental charge',
        'quantity': 'charge',
    },
    {
        'symbol': 'V',
        'name': 'volt',
        'quantity': 'potential',
    },
    {
        'symbol': 'Ω',
        'name': 'ohm',
        'quantity': 'resistance',
    },
    {
        'symbol': 'S',
        'name': 'seimens',
        'quantity': 'conductance',
    },
    {
        'symbol': 'F',
        'name': 'farad',
        'quantity': 'capacitance',
    },
    {
        'symbol': 'Wb',
        'name': 'weber',
        'quantity': 'magnetic flux',
    },
    {
        'symbol': 'Mx',
        'name': 'maxwell',
        'quantity': 'magnetic flux',
    },
    {
        'symbol': 'Oe',
        'name': 'Oersted',
        'quantity': 'magnetic intensity',
    },
    {
        'symbol': 'H',
        'name': 'henry',
        'quantity': 'inductance',
    },
    {
        'symbol': 'T',
        'name': 'tesla',
        'quantity': 'magnetic induction',
    },
    {
        'symbol': 'G',
        'name': 'gauss',
        'quantity': 'magnetic induction',
    },
    {
        'symbol': 'lm',
        'name': 'lumen',
        'quantity': 'luminous flux',
    },
    {
        'symbol': 'lx',
        'name': 'lux',
        'quantity': 'illuminance',
    },
    {
        'symbol': 'Bq',
        'name': 'becquerel',
        'quantity': 'radioactivity',
    },
    {
        'symbol': 'Ci',
        'name': 'Curie',
        'quantity': 'radioactivity',
    },
    {
        'symbol': 'Gy',
        'name': 'gray',
        'quantity': 'dosage',
    },
    {
        'symbol': 'P',
        'name': 'poise',
        'quantity': 'viscosity',
    },
    {
        'symbol': '1',
        'name': 'unitless',
        'quantity': 'identity',
    },
]


def _build_named_units(
    prefixes: typing.List[typing.Dict[T, typing.Any]],
    units: typing.List[typing.Dict[T, str]],
) -> aliased.Mapping[T, typing.Dict[T, typing.Any]]:
    """Helper for building the collection of named units."""
    # Find the null metric prefix (0).
    null_prefix = next(
        prefix for prefix in prefixes if prefix['factor'] == 0.0
    )
    # Find the metric prefix corresponding to unity (1).
    base_prefix = next(
        prefix for prefix in prefixes if prefix['factor'] == 1.0
    )
    # Find the unitless quantity.
    unitless = next(
        unit for unit in units if unit['quantity'] == 'identity'
    )
    # Build mapping from (prefix, unit) combinations, as appropriate.
    mapped = {}
    for unit in units:
        if unit is unitless:
            # Prevent meaningless unitless expressions (e.g., cenitunitless).
            mapped[(unitless['symbol'], unitless['name'])] = {
                'prefix': base_prefix, 'base': unitless
            }
        else:
            for prefix in prefixes:
                if prefix is base_prefix:
                    # Add the unscaled unit.
                    mapped[(unit['symbol'], unit['name'])] = {
                        'prefix': base_prefix, 'base': unit
                    }
                elif prefix is not null_prefix:
                    # Add the full (prefix, unit) combination.
                    key = [
                        f"{prefix['symbol']}{unit['symbol']}",
                        f"{prefix['name']}{unit['name']}"
                    ]
                    if prefix['symbol'] == 'μ':
                        key += [f"u{unit['symbol']}"]
                    mapped[tuple(key)] = {'prefix': prefix, 'base': unit}
    return aliased.Mapping(mapped)


NAMED_UNITS = _build_named_units(_PREFIXES, _UNITS)


# A note about angles: Kalinin (2019) "On the status of plane and solid in the
# International System of Units (SI)" makes a very compelling argument that the
# plane- and solid-angle units should not be '1'; instead:
# - the plane-angle unit should be 'radian' ('rad'),
# - the solid-angle unit should be 'steradian' ('sr'), and
# - the plane angle should be a base quantity,
# - 'radian' should be considered a base unit,
# - 'steradian' should be considered a derived unit, with 'sr = rad^2'.

# References and notes on quantities, dimensions, and units:
# - https://en.wikipedia.org/wiki/International_System_of_Quantities#Base_quantities
# - https://www.nist.gov/pml/weights-and-measures/metric-si/si-units
# - This module uses 'H' to represent the dimension of temperature because the
#   SI character, $\Theta$, is not an ASCII character. I also considered 'O'
#   because of its similarity to $\Theta$, but it looks too much like '0'
#   (zero), which, ironically, looks more like $\Theta$ than 'O'.
# - This module adds an identity quantity with '1' as its dimension and unit

_BASE_QUANTITIES = container.Table(
    [
        {'name': 'identity', 'dimension': '1', 'unit': '1'},
        {'name': 'amount', 'dimension': 'N', 'unit': 'mol'},
        {'name': 'current', 'dimension': 'I', 'unit': 'A'},
        {'name': 'length', 'dimension': 'L', 'unit': 'm'},
        {'name': 'luminous intensity', 'dimension': 'J', 'unit': 'cd'},
        {'name': 'mass', 'dimension': 'M', 'unit': 'kg'},
        {'name': 'temperature', 'dimension': 'H', 'unit': 'K'},
        {'name': 'time', 'dimension': 'T', 'unit': 's'},
    ]
)

_QUANTITIES = {
    'acceleration': 'velocity / time',
    'amount': {
        'dimensions': {
            'mks': 'N',
            'cgs': 'N',
        },
        'units': {
            'mks': 'mol',
            'cgs': 'mol',
        },
    },
    'area': 'length^2',
    'capacitance': {
        'dimensions': {
            'mks': '(T^2 * I)^2 / (M * L^2)',
            'cgs': 'L',
        },
        'units': {
            'mks': 'F',
            'cgs': 'cm',
        },
    },
    'charge': {
        'dimensions': {
            'mks': 'I * T',
            'cgs': '(M^1/2 * L^3/2) / T',
        },
        'units': {
            'mks': 'C',
            'cgs': 'statC',
        },
    },
    'charge density': 'charge / volume',
    'conductance': {
        'dimensions': {
            'mks': '(T^3 * I^2) / (M * L^2)',
            'cgs': 'L / T',
        },
        'units': {
            'mks': 'S',
            'cgs': 'cm / s',
        },
    },
    'conductivity': 'conductance / length',
    'current': {
        'dimensions': {
            'mks': 'I',
            'cgs': '(M^1/2 * L^3/2) / T^2',
        },
        'units': {
            'mks': 'A',
            'cgs': 'statA',
        },
    },
    'current density': 'current / area',
    'displacement': {
        'dimensions': {
            'mks': 'I * T / L^2',
            'cgs': 'M^1/2 / (L^1/2 * T)',
        },
        'units': {
            'mks': 'C / m^2',
            'cgs': 'statC / m^2',
        },
    },
    'dosage': {
        'dimensions': {
            'mks': 'L^2 / T^2',
            'cgs': 'L^2 / T^2',
        },
        'units': {
            'mks': 'Gy',
            'cgs': 'erg / g', # Historically, 'rad', but that's in use.
        },
    },
    'electric charge': 'charge',
    'electric field': 'potential / length',
    'electromotance': 'potential',
    'energy': {
        'dimensions': {
            'mks': '(M * L^2) / T^2',
            'cgs': '(M * L^2) / T^2',
        },
        'units': {
            'mks': 'J',
            'cgs': 'erg',
        },
    },
    'energy density': 'energy / volume',
    'fluence': 'particle fluence',
    'flux': 'particle flux',
    'force': {
        'dimensions': {
            'mks': '(M * L) / T^2',
            'cgs': '(M * L) / T^2',
        },
        'units': {
            'mks': 'N',
            'cgs': 'dyn',
        },
    },
    'frequency': {
        'dimensions': {
            'mks': '1 / T',
            'cgs': '1 / T',
        },
        'units': {
            'mks': 'Hz',
            'cgs': 'Hz',
        },
    },
    'identity': {
        'dimensions': {
            'mks': '1',
            'cgs': '1',
        },
        'units': {
            'mks': '1',
            'cgs': '1',
        },
    },
    'illumunance': { # See note about radian (Kalinin 2019).
        'dimensions': {
            'mks': 'J / L^2',
            'cgs': 'J / L^2',
        },
        'units': {
            'mks': 'cd * sr / m^2',
            'cgs': 'cd * sr / cm^2',
        },
    },
    'impedance': {
        'dimensions': {
            'mks': '(M * L^2) / (T^3 * I)',
            'cgs': 'T / L',
        },
        'units': {
            'mks': 'ohm',
            'cgs': 's / cm',
        },
    },
    'inductance': {
        'dimensions': {
            'mks': '(M * L^2) / (I * T)^2',
            'cgs': 'T^2 / L',
        },
        'units': {
            'mks': 'H',
            'cgs': 's^2 / cm',
        },
    },
    'induction': 'magnetic induction',
    'integral flux': 'flux * energy',
    'length': {
        'dimensions': {
            'mks': 'L',
            'cgs': 'L',
        },
        'units': {
            'mks': 'm',
            'cgs': 'cm',
        },
    },
    'luminous flux': { # See note about radian (Kalinin 2019).
        'dimensions': {
            'mks': 'J',
            'cgs': 'J',
        },
        'units': {
            'mks': 'cd * sr',
            'cgs': 'cd * sr',
        },
    },
    'luminous intensity': {
        'dimensions': {
            'mks': 'J',
            'cgs': 'J',
        },
        'units': {
            'mks': 'cd',
            'cgs': 'cd',
        },
    },
    'magnetic field': 'magnetic induction',
    'magnetic flux': {
        'dimensions': {
            'mks': '(M * L^2) / (T^2 * I)',
            'cgs': '(M^1/2 * L^3/2) / T',
        },
        'units': {
            'mks': 'Wb',
            'cgs': 'Mx',
        },
    },
    'magnetic induction': {
        'dimensions': {
            'mks': 'M / (T^2 * I)',
            'cgs': 'M^1/2 / (L^1/2 * T)',
        },
        'units': {
            'mks': 'T',
            'cgs': 'G',
        },
    },
    'magnetic intensity': {
        'dimensions': {
            'mks': 'I / L',
            'cgs': 'M^1/2 / (L^1/2 * T)',
        },
        'units': {
            'mks': 'A / m',
            'cgs': 'Oe',
        },
    },
    'magnetic moment': {
        'dimensions': {
            'mks': 'I * L^2',
            'cgs': '(M^1/2 * L^5/2) / T',
        },
        'units': {
            'mks': 'A * m^2',
            'cgs': 'Oe * cm^3',
        },
    },
    'magnetization': 'magnetic intensity',
    'magnetomotance': 'current',
    'mass': {
        'dimensions': {
            'mks': 'M',
            'cgs': 'M',
        },
        'units': {
            'mks': 'kg',
            'cgs': 'g',
        },
    },
    'mass density': 'mass / volume',
    'mass number': 'number',
    'momentum': {
        'dimensions': {
            'mks': '(M * L) / T',
            'cgs': '(M * L) / T',
        },
        'units': {
            'mks': 'kg * m / s',
            'cgs': 'g * cm / s',
        },
    },
    'momentum density': 'momentum / volume',
    'number': 'identity',
    'number density': '1 / volume',
    'particle distribution': '1 / (length * velocity)^3',
    'particle fluence': 'number / (area * solid_angle * energy / mass_number)',
    'particle flux': 'fluence / time',
    'permeability': {
        'dimensions': {
            'mks': '(M * L) / (I * T)^2',
            'cgs': '1',
        },
        'units': {
            'mks': 'H / m',
            'cgs': '1',
        },
    },
    'permittivity': {
        'dimensions': {
            'mks': 'T^4 * I / (M * L^3)',
            'cgs': '1',
        },
        'units': {
            'mks': 'F / m',
            'cgs': '1',
        },
    },
    'plane angle': { # See note about radian (Kalinin 2019).
        'dimensions': {
            'mks': '1',
            'cgs': '1',
        },
        'units': {
            'mks': 'rad',
            'cgs': 'rad',
        },
    },
    'polarization': 'charge / area',
    'potential': {
        'dimensions': {
            'mks': '(M * L^2) / (T^3 * I)',
            'cgs': '(M^1/2 * L^1/2) / T',
        },
        'units': {
            'mks': 'V',
            'cgs': 'statV',
        },
    },
    'power': {
        'dimensions': {
            'mks': 'M * L^2 / T^3',
            'cgs': 'M * L^2 / T^3',
        },
        'units': {
            'mks': 'W',
            'cgs': 'erg / s',
        },
    },
    'power density': 'power / volume',
    'pressure': {
        'dimensions': {
            'mks': 'M / (L * T^2)',
            'cgs': 'M / (L * T^2)',
        },
        'units': {
            'mks': 'Pa',
            'cgs': 'dyn / cm^2', # also barye (Ba)?
        },
    },
    'radioactivity': {
        'dimensions': {
            'mks': '1 / T',
            'cgs': '1 / T',
        },
        'units': {
            'mks': 'Bq',
            'cgs': 'Ci',
        },
    },
    'rate': 'frequency',
    'ratio': 'identity',
    'reluctance': {
        'dimensions': {
            'mks': '(I * T)^2 / (M * L^2)',
            'cgs': '1 / L',
        },
        'units': {
            'mks': 'A / Wb',
            'cgs': '1 / cm',
        },
    },
    'resistance': 'impedance',
    'resistivity': 'resistance * length',
    'temperature': {
        'dimensions': {
            'mks': 'H',
            'cgs': 'H',
        },
        'units': {
            'mks': 'K',
            'cgs': 'K',
        },
    },
    'thermal conductivity': 'power / (length * temperature)',
    'time': {
        'dimensions': {
            'mks': 'T',
            'cgs': 'T',
        },
        'units': {
            'mks': 's',
            'cgs': 's',
        },
    },
    'solid angle': { # See note about radian (Kalinin 2019).
        'dimensions': {
            'mks': '1',
            'cgs': '1',
        },
        'units': {
            'mks': 'sr',
            'cgs': 'sr',
        },
    },
    'speed': 'velocity',
    'vector potential': {
        'dimensions': {
            'mks': '(M * L) / (T^2 * I)',
            'cgs': '(M^1/2 * L^1/2) / T',
        },
        'units': {
            'mks': 'Wb / m',
            'cgs': 'G * cm',
        },
    },
    'velocity': 'length / time',
    'viscosity': {
        'dimensions': {
            'mks': 'M / (L * T)',
            'cgs': 'M / (L * T)',
        },
        'units': {
            'mks': 'kg / (m * s)',
            'cgs': 'P',
        },
    },
    'volume': 'length^3',
    'vorticity': 'frequency',
    'wavenumber': '1 / length',
    'work': 'energy',
}


C = 2.99792458e10
"""The speed of light in cm/s."""
PI = numpy.pi
"""The ratio of a circle's circumference to its diameter."""


_CONVERSIONS = {
    ('Rs', 'm'): 6.96e8,
    ('F', 'cm'): C**2 * 1e-9,
    ('C', 'statC'): 10*C,
    ('e', 'C'): 1.6022e-19,
    ('S', 'cm / s'): C**2 * 1e-5,
    ('A', 'statA'): 10*C,
    # ('C / m^2', 'statC / m^2'): 4*PI * C * 1e-3,
    ('Gy', 'erg / g'): 1e4,
    ('J', 'erg'): 1e7,
    ('eV', 'J'): 1.6022e-19,
    ('N', 'dyn'): 1e5,
    ('ohm', 's / cm'): 1e5 / C**2,
    ('H', 's^2 / cm'): 1e5 / C**2,
    # ('m', 'cm'): 1e2,
    ('au', 'm'): 1.495978707e11,
    ('Wb', 'Mx'): 1e8,
    ('T', 'G'): 1e4,
    ('A / m', 'Oe'): 4*PI * 1e-3,
    # ('A * m^2', 'Oe * cm^3'): 1e-3,
    # ('kg', 'g'): 1e3,
    ('nuc', 'kg'): 1.6605e-27,
    ('amu', 'kg'): 1.6605e-27,
    # ('kg * m / s', 'g * cm / s'): 1e5,
    ('H / m', '1'): 1e7 / 4*PI,
    ('F / m', '1'): 36*PI * 1e9,
    ('rad', 'deg'): 180 / PI,
    ('V', 'statV'): 1e6 / C,
    ('W', 'erg / s'): 1e7,
    ('Pa', 'dyn / cm^2'): 1e1,
    ('Bq', 'Ci'): 1.0 / 3.7e10,
    ('A / Wb', '1 / cm'): 4*PI * 1e-9,
    ('s', 'min'): 1.0 / 60.0,
    ('s', 'h'): 1.0 / 3600.0,
    ('s', 'd'): 1.0 / 86400.0,
    ('Wb / m', 'G * cm'): 1e6,
    ('kg / (m * s)', 'P'): 1e1,
}


CONVERSIONS = container.Graph(_CONVERSIONS)


_NONSTANDARD = {
    'julian date': 'day',
    'shell': '1',
    'cos(mu)': '1',
    'e-': 'e',
    '# / cm^2 s sr MeV': '# / cm^2 s sr MeV/nuc',
}
"""Internal substitutions for non-standard EPREM units.

This dictionary maps each non-standard unit associated with an EPREM output
quantity to a unit defined in `~metric.py`.
"""

def standardize(unit: T):
    """Replace this unit string with a standard unit string, if possible.

    This function looks for `unit` in the known conversions and returns the
    standard unit string if it exists. If this doesn't find a standard unit
    string, it just returns the input.
    """
    unit = _NONSTANDARD.get(str(unit), unit)
    if '/' in unit:
        num, den = str(unit).split('/', maxsplit=1)
        unit = ' / '.join((num.strip(), f"({den.strip()})"))
    return unit




import abc
import typing

import numpy

from .. import base
from .. import container
from .. import numeric
from ._types import ValueType


T = typing.TypeVar('T')


class Variable(numeric.Object[ValueType], base.abc.Sequence):
    """A sequence-like object with conditional cast to built-in type.
    
    An instance of this class behaves like a sequence of real values, and
    additionally supports converting single-valued instances to built-in `int`,
    `float`, or `complex` type.

    See Also
    --------
    `~quantity.abc.Sequence`
    """

    def __init__(self, data: typing.Sequence[ValueType]) -> None:
        super().__init__(data)
        self._size = None

    def __int__(self) -> int:
        """Called for int(self)."""
        return self._cast_to(int)

    def __float__(self) -> float:
        """Called for float(self)."""
        return self._cast_to(float)

    def __complex__(self) -> complex:
        """Called for complex(self)."""
        return self._cast_to(complex)

    def _cast_to(self, __type: typing.Type[T]) -> T:
        """Internal logic for casting data to type `T`."""
        if self.size == 1:
            return __type(self.data[0])
        raise TypeError(
            f"Cannot convert size-{self.size} object to float"
        ) from None

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, i: typing.SupportsIndex):
        return self.data[i]

    def __array__(self, *args, **kwargs) -> numpy.ndarray:
        """Called to convert this variable to a `numpy.ndarray`.
        
        Notes
        -----
        - This method converts an instance to a `numpy.ndarray` by calling
          `numpy.asarray` on the underlying data. Therefore, if the data is
          already a `numpy.ndarray`, calling `numpy.asarray` on this instance
          will return the original array.
        """
        return numpy.asarray(self.data, *args, **kwargs)

    @property
    def size(self) -> int:
        """The number of numerical values in this instance."""
        if self._size is None:
            self._size = container.size(self._data)
        return self._size


class Quantity(base.abc.Real):
    """Partial implementation of the real-valued protocol."""

    # NOTE: The body of this class defines all real-valued methods in order to
    # add type annotations. Many of these definitions simply call back to the
    # parent ABC and therefore remain abstract.

    # (abstract)
    def __eq__(self: base.RealType, other) -> bool:
        return super().__eq__(other)

    # (concrete)
    def __ne__(self: base.RealType, other) -> bool:
        return not (self == other)

    # (abstract)
    def __lt__(self: base.RealType, other) -> bool:
        return super().__lt__(other)

    # (concrete)
    def __le__(self: base.RealType, other) -> bool:
        return self < other or self == other

    # (concrete)
    def __gt__(self: base.RealType, other) -> bool:
        return not (self < other or self == other)

    # (concrete)
    def __ge__(self: base.RealType, other) -> bool:
        return not (self < other)

    # (abstract)
    def __abs__(self: base.RealType) -> base.RealType:
        return super().__abs__()

    # (abstract)
    def __pos__(self: base.RealType) -> base.RealType:
        return super().__pos__()

    # (abstract)
    def __neg__(self: base.RealType) -> base.RealType:
        return super().__neg__()

    # (abstract)
    def __add__(self: base.RealType, other) -> base.RealType:
        return super().__add__(other)

    # (abstract)
    def __radd__(self: base.RealType, other) -> base.RealType:
        return super().__radd__(other)

    # (concrete)
    def __sub__(self: base.RealType, other) -> base.RealType:
        return self + (-other)

    # (concrete)
    def __rsub__(self: base.RealType, other) -> base.RealType:
        return other + (-self)

    # (abstract)
    def __mul__(self: base.RealType, other) -> base.RealType:
        return super().__mul__(other)

    # (abstract)
    def __rmul__(self: base.RealType, other) -> base.RealType:
        return super().__rmul__(other)

    # (abstract)
    def __truediv__(self: base.RealType, other) -> base.RealType:
        return super().__truediv__(other)

    # (abstract)
    def __rtruediv__(self: base.RealType, other) -> base.RealType:
        return super().__rtruediv__(other)

    # (abstract)
    def __pow__(self: base.RealType, other) -> base.RealType:
        return super().__pow__(other)

    @abc.abstractmethod
    def __floordiv__(self, other):
        """Called for self // other."""

    @abc.abstractmethod
    def __rfloordiv__(self, other):
        """Called for other // self."""

    @abc.abstractmethod
    def __mod__(self, other):
        """Called for self % other."""

    @abc.abstractmethod
    def __rmod__(self, other):
        """Called for other % self."""





import typing

from .. import base
from .. import metric
from .. import measured
from .. import real
from ._functions import (
    isindexlike,
    ismeasurable,
    parse,
    hasdata,
    getdata,
)
from ._exceptions import (
    MeasuringTypeError,
    ParsingTypeError,
    ParsingValueError,
)
from ._measurement import Measurement


__all__ = [
    'isindexlike',
    'ismeasurable',
    'measure',
    'parse',
    'hasdata',
    'getdata',
    'MeasuringTypeError',
    'ParsingTypeError',
    'ParsingValueError',
    'Measurement',
]


@typing.overload
def measure(*args) -> Measurement:
    """Measure the given arguments, if possible.

    Parameters
    ----------
    *args
        One or more objects from which to create a measurement. See
        `~ismeasurable` for a description of supported object types.

    Returns
    -------
    `~Measurement`

    Raises
    ------
    `~MeasuringTypeError`
        The input was empty or the parsing algorithm was not able to determine
        appropriate numeric values and a unique unit from the input arguments.

    See Also
    --------
    `~measurable.parse`
        The function that parses measurable input into numeric values and an
        associated unit, if possible.
    `~measurable.ismeasurable`
        The function that detects measurable input.
    """

@typing.overload
def measure(x: measured.Object) -> Measurement:
    """Measure an existing measured object.
    
    Parameters
    ----------
    x : `~measured.Object`
        The measured object to measure. See Notes for additional information
        about how this function handles specific types of measured objects.

    Returns
    -------
    `~Measurement`

    Raises
    ------
    `~MeasurementTypeError`
        The type of object returned by `x.__measure__` was not an instance of
        `~Measurement`.

    Notes
    -----
    - If `x` is already an instance of `~Measurement`, this function will
      immediately return it.
    - If `x` implements the special method `__measure__`, this function will
      defer to that implementation.
    - If `x` is any other instance of `~measured.Object` or one of its
      subclasses, this function will create a new measurement from the
      real-valued data in `x.data` and the metric unit in `x.unit`.
    """


@typing.overload
def measure(x: base.MeasurableType, **kwargs) -> Measurement:
    """Measure an explicitly measurable object.
    
    Parameters
    ----------
    x
        An object that complies with the `~quantity.Measurable` protocol by
        implementing the special method `__measure__`.

    **kwargs
        Keyword arguments to pass to `x.__measure__`.

    Returns
    -------
    `~Measurement`

    Raises
    ------
    `~MeasurementTypeError`
        The type of object returned by `x.__measure__` was not an instance of
        `~Measurement`.
    """

def measure(*args, **kwargs):
    if not args:
        raise MeasuringTypeError("There is nothing to measure") from None
    this = args[0]
    if isinstance(this, Measurement):
        return this
    if isinstance(this, base.Measurable):
        return _measure_explicit(this, **kwargs)
    if isinstance(this, measured.Object):
        return measurement(this.data, unit=this.unit)
    return _measure_implicit(args)


def _measure_explicit(x: base.Measurable, **kwargs):
    """Create a measurement by calling `x.__measure__`."""
    result = x.__measure__(**kwargs)
    if not isinstance(result, (Measurement, base.Measurement)):
        raise MeasuringTypeError(
            f"{type(x)}.__measure__ returned"
            f" non-{Measurement} (type {type(result)})"
        )
    return result


def _measure_implicit(x):
    """Create a measurement by calling `metric.parse`."""
    try:
        parsed = parse(x, distribute=False)
    except (ParsingValueError, ParsingTypeError) as err:
        raise MeasuringTypeError(f"Cannot measure {x}") from err
    data = parsed[:-1]
    unit = parsed[-1]
    return measurement(data, unit=unit)


@typing.overload
def measurement(
    x: typing.Union[real.ValueType, typing.Sequence[real.ValueType]],
    /,
    unit: metric.Unit=typing.Literal['1']
) -> Measurement[real.ValueType]:
    """Create a physical measurement.
    
    Parameters
    ----------
    data : real number or sequence of real numbers
        The numerical value(s) from which to create a new measurement.

    unit : `~metric.Unit`
        The metric unit associated with `data`.

    Returns
    -------
    `~Measurement`

    Raises
    ------
    `TypeError`
        The caller passed a single non-numeric object or a sequence that
        includes non-numeric members.
    """

def measurement(*args, **kwargs):
    return _measurement.measurement_factory(*args, **kwargs)





"""
Support for working with EPREM simulation parameters.
"""

import typing

from .. import metric
from .. import paths
from .. import physical
from .. import quantity
from .. import real
from ._config import (
    ConfigFile,
    configfile_factory as configfile,
    Defaults,
    defaults_factory as defaults,
)
from ._metadata import (
    ALIASES,
)
from ._runtime import (
    Interface,
    Variable,
)
from ._src import (
    BaseTypesH,
    ConfigurationC,
    SourceFile,
)
from . import reference


DB_PATH = str(SourceFile._db_path)
"""Public string version of source file default database path."""


__all__ = [
    "ALIASES",
    "BaseTypesH",
    "ConfigurationC",
    "ConfigFile",
    "Defaults",
    "Interface",
    "SourceFile",
    "Variable",
    "configfile",
    "defaults",
    "reference",
    "interface",
    "variable",
]


ArgumentType = typing.Union[physical.Scalar, physical.Tensor]


@typing.overload
def variable(
    x: typing.Any,
    /,
    unit: typing.Optional[metric.UnitLike],
) -> _runtime.Variable:
    """Create a variable from the given arguments, if possible"""

@typing.overload
def variable(
    x: typing.Sequence[real.ValueType],
    /,
) -> _runtime.Variable[real.ValueType]:
    """Create a unitless variable from a real-valued sequence.

    Parameters
    ----------
    x : sequence of `int`, `float`, or `numbers.Real`
        The numeric value(s) of the new variable. This function will convert the
        given sequence to a `numpy.ndarray` if necessary.

    Returns
    -------
    `~Variable`
        A new variable with the given value(s), and a unit equal to
        '1'.

    Raises
    ------
    `TypeError`
        The input is not logically 1-dimensional.
    """

@typing.overload
def variable(
    x: typing.Sequence[real.ValueType],
    /,
    unit: metric.UnitLike,
) -> _runtime.Variable[real.ValueType]:
    """Create a variable from a real-valued sequence and a unit.

    Parameters
    ----------
    x : sequence of `int`, `float`, or `numbers.Real`
        The numeric value(s) of the new variable. This function will convert the
        given sequence to a `numpy.ndarray` if necessary.
    unit : unit-like
        The metric unit associated with `x`.

    Returns
    -------
    `~Variable`
        A new variable with the given value(s) and unit.

    Raises
    ------
    `TypeError`
        The input is not logically 1-dimensional.
    """

@typing.overload
def variable(
    x: _runtime.Variable[real.ValueType],
    /,
) -> _runtime.Variable[real.ValueType]:
    """Create a variable from an existing variable.

    Parameters
    ----------
    x : `~Variable`
        The variable from which to create a new variable. This function will
        make a copy of `x`.

    Returns
    -------
    `~Variable`
        A new variable with value and unit taken from the given
        variable.

    Raises
    ------
    `ValueError`
        The caller passed an explicit unit.
    """

@typing.overload
def variable(
    x: physical.Scalar[real.ValueType],
    /,
) -> _runtime.Variable[real.ValueType]:
    """Create a variable from a physical scalar.

    Parameters
    ----------
    x : `~physical.Scalar`
        The scalar from which to create a new variable. This function will
        convert the scalar's data value into a 1-element `numpy.ndarray`.

    Returns
    -------
    `~Variable`
        A new variable with value and unit taken from the given scalar.

    Raises
    ------
    `ValueError`
        The caller passed an explicit unit.
    """

@typing.overload
def variable(
    x: physical.Vector[real.ValueType],
    /,
) -> _runtime.Variable[real.ValueType]:
    """Create a variable from a physical vector.

    Parameters
    ----------
    x : `~physical.Vector`
        The vector from which to create a new variable.

    Returns
    -------
    `~Variable`
        A new variable with value and unit taken from the given vector.

    Raises
    ------
    `ValueError`
        The caller passed an explicit unit.
    """

@typing.overload
def variable(
    x: quantity.Measurement[real.ValueType],
    /,
) -> _runtime.Variable[real.ValueType]:
    """Create a variable from an existing measurement.

    Parameters
    ----------
    x : `~quantity.Measurement`
        The measurement from which to create a new variable. This function will
        convert the measurement's data into a `numpy.ndarray` if necessary.

    Returns
    -------
    `~Variable`
        A new variable with data and unit taken from the given
        measurement.

    Raises
    ------
    `ValueError`
        The caller passed an explicit unit.
    """

def variable(x, /, unit=None):
    return _runtime.variable_factory(x, unit=unit)


@typing.overload
def interface(
    config: typing.Optional[paths.PathLike]=None,
    srcdir: typing.Optional[paths.PathLike]=None,
    **options
) -> Interface:
    """Create an interface to EPREM parameter values.

    Parameters
    ----------
    config : path-like, optional
        The location of the EPREM configuration file containing user-provided
        parameter values for a particular EPREM simulation run. May be omitted,
        in which case the interface will use default argument values.

    srcdir : path-like, optional
        The directory containing EPREM `configuration.c`. May be omitted, in
        which case the interface will use pre-computed values.

    **options
        Options for loading the configuration file. See `~ConfigFile`
        for more information.

    Notes
    -----
    - Path-like arguments may be any path-like object (e.g., `str`,
      `pathlib.Path`), may be relative, and may contain standard wildcard
      characters (e.g., `~`).
    """

def interface(*args, **kwargs):
    return _runtime.interface_factory(*args, **kwargs)





import pathlib
import typing

import numpy.typing

from .. import atomic
from .. import etc
from .. import metric
from .. import paths
from .. import physical
from ._reference import AXES
from ._viewers import view_factory


@etc.autostr
class Axes(typing.Mapping[str, physical.Axis]):
    """A interface to the canonical array axes."""

    def __init__(
        self,
        source: pathlib.Path,
        system: metric.System,
    ) -> None:
        """Initialize this interface."""
        self._source = source
        self._system = system
        self._view = None
        self._time = None
        self._shell = None
        self._species = None
        self._energy = None
        self._mu = None

    @typing.overload
    def __getitem__(
        self,
        __k: typing.Literal['time'],
    ) -> physical.Coordinates: ...

    @typing.overload
    def __getitem__(
        self,
        __k: typing.Literal['shell'],
    ) -> physical.Points: ...

    @typing.overload
    def __getitem__(
        self,
        __k: typing.Literal['species'],
    ) -> physical.Symbols: ...

    @typing.overload
    def __getitem__(
        self,
        __k: typing.Literal['energy'],
    ) -> physical.Coordinates: ...

    @typing.overload
    def __getitem__(
        self,
        __k: typing.Literal['mu'],
    ) -> physical.Coordinates: ...

    @typing.overload
    def __getitem__(self, __k: str) -> physical.Axis: ...

    def __getitem__(self, __k: str):
        """Called for key-based access."""
        try:
            axis = getattr(self, __k)
        except AttributeError as err:
            raise KeyError(
                f"No axis named {__k!r} in {self.view!r}"
            ) from err
        return axis

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(AXES)

    def __iter__(self) -> typing.Iterator[str]:
        """Called for iter(self)."""
        return iter(AXES)

    @property
    def time(self):
        """The output times."""
        if self._time is None:
            data = self._create('time')
            self._time = physical.coordinates(data)
        return self._time

    @property
    def shell(self):
        """The simulation node layers."""
        if self._shell is None:
            data = self.view.arrays['shell'].data
            self._shell = physical.points(data)
        return self._shell

    @property
    def species(self):
        """The energetic-particle species."""
        if self._species is None:
            mass = self._create('mass', unit='nuc')
            charge = self._create('charge', unit='e')
            data = atomic.elements(mass=mass, charge=charge)
            self._species = physical.symbols(data)
        return self._species

    @property
    def energy(self) -> physical.Array:
        """The particle energies."""
        if self._energy is None:
            data = numpy.squeeze(self._create('egrid'))
            self._energy = physical.coordinates(data)
        return self._energy

    @property
    def mu(self):
        """The particle pitch-angle cosines."""
        if self._mu is None:
            data = self._create('mu')
            self._mu = physical.coordinates(data)
        return self._mu

    def _create(
        self,
        name: str,
        unit: metric.UnitLike=None,
    ) -> physical.Tensor:
        """Create an array for the named quantity."""
        try:
            base = self.view.arrays[name]
        except KeyError as err:
            raise KeyError(f"No quantity {name!r} in {self.view!r}") from err
        array = physical.tensor(
            numpy.array(base.data),
            unit=metric.standardize(base.unit),
        )
        return array.withunit(unit or str(self.system))

    @property
    def view(self):
        """A view into this dataset on disk."""
        if self._view is None:
            self._view = view_factory(self.source)
        return self._view
    
    @property
    def source(self):
        """This dataset's source path."""
        return self._source

    @property
    def system(self):
        """The metric system in which to represent this dataset."""
        return self._system


@etc.autostr
class Arrays:
    """An interface to array-like quantities in a fixed metric system."""

    def __init__(
        self,
        source: pathlib.Path,
        system: metric.System,
    ) -> None:
        """Initialize this interface."""
        self._source = source
        self._system = system
        self._axes = None
        self._view = None
        self._time = None
        self._shell = None
        self._species = None
        self._energy = None
        self._mu = None
        self._v = None
        self._mass = None
        self._charge = None
        self._r = None
        self._theta = None
        self._phi = None
        self._br = None
        self._btheta = None
        self._bphi = None
        self._ur = None
        self._utheta = None
        self._uphi = None
        self._rho = None
        self._mfp = None
        self._f = None
        self._flux = None
        self._hasmfp = None
        self._hasdist = None
        self._hasflux = None

    @property
    def time(self):
        """The output times."""
        if self._time is None:
            self._time = self._create('time')
        return self._time

    @property
    def shell(self):
        """The simulation node layers."""
        if self._shell is None:
            self._shell = self.view.arrays['shell'].data
        return self._shell

    @property
    def species(self):
        """The energetic-particle species."""
        if self._species is None:
            mass = self._create('mass', unit='nuc')
            charge = self._create('charge', unit='e')
            self._species = atomic.elements(mass=mass, charge=charge)
        return self._species

    @property
    def energy(self) -> physical.Array:
        """The particle energies."""
        if self._energy is None:
            self._energy = numpy.squeeze(self._create('egrid'))
        return self._energy

    @property
    def v(self) -> physical.Array:
        """The particle speeds."""
        if self._v is None:
            self._v = numpy.squeeze(self._create('vgrid'))
        return self._v

    @property
    def mu(self):
        """The particle pitch-angle cosines."""
        if self._mu is None:
            self._mu = self._create('mu')
        return self._mu

    @property
    def mass(self):
        """The mass of each species."""
        if self._mass is None:
            self._mass = self._create('mass', unit=self.system['mass'].unit)
        return self._mass

    @property
    def charge(self):
        """The charge of each species."""
        if self._charge is None:
            self._charge = self._create(
                'charge',
                unit=self.system['charge'].unit,
            )
        return self._charge

    @property
    def r(self):
        """The radial observer coordinates."""
        if self._r is None:
            self._r = self._create('R')
        return self._r

    @property
    def theta(self):
        """The polar observer coordinates."""
        if self._theta is None:
            self._theta = self._create('T')
        return self._theta

    @property
    def phi(self):
        """The azimuthal observer coordinates."""
        if self._phi is None:
            self._phi = self._create('P')
        return self._phi

    @property
    def br(self):
        """The magnetic-field radial component."""
        if self._br is None:
            self._br = self._create('Br')
        return self._br

    @property
    def btheta(self):
        """The magnetic-field polar component."""
        if self._btheta is None:
            self._btheta = self._create('Bt')
        return self._btheta

    @property
    def bphi(self):
        """The magnetic-field polar component."""
        if self._bphi is None:
            self._bphi = self._create('Bp')
        return self._bphi

    @property
    def ur(self):
        """The velocity-field radial component."""
        if self._ur is None:
            self._ur = self._create('Vr')
        return self._ur

    @property
    def utheta(self):
        """The velocity-field polar component."""
        if self._utheta is None:
            self._utheta = self._create('Vt')
        return self._utheta

    @property
    def uphi(self):
        """The velocity-field polar component."""
        if self._uphi is None:
            self._uphi = self._create('Vp')
        return self._uphi

    @property
    def rho(self):
        """The plasma density."""
        if self._rho is None:
            self._rho = self._create('Rho')
        return self._rho

    @property
    def mfp(self):
        """The particle mean free path."""
        if self._mfp is None:
            self._mfp = self._create('mfp')
        return self._mfp

    @property
    def f(self):
        """The particle distribution."""
        if self._f is None:
            if not self.hasflux:
                self._f = self._create('Dist')
        return self._f

    @property
    def flux(self):
        """The particle flux."""
        if self._flux is None:
            if self.hasflux:
                self._flux = self._create('flux')
        return self._flux

    def _create(
        self,
        name: str,
        unit: metric.UnitLike=None,
    ) -> physical.Array:
        """Create an array for the named quantity."""
        try:
            base = self.view.arrays[name]
        except KeyError as err:
            raise KeyError(f"No observable {name!r} in {self.view!r}") from err
        axes = {d: self.axes[d] for d in base.dimensions}
        array = physical.array(
            base.data,
            unit=metric.standardize(base.unit),
            axes=physical.axes(axes),
        )
        return array.withunit(unit or str(self.system))

    @property
    def hasmfp(self):
        """True if this dataset contains the particle mean free path.
        
        Datasets from EPREM runs prior to 11 Mar 2025 did not contain the
        particle mean free path. This property allows users to check whether the
        underlying dataset contains the mean free path (`True`) or not (`False`)
        before attempting to access the corresponding array.
        """
        if self._hasmfp is None:
            self._hasmfp = 'mfp' in self.view.arrays
        return self._hasmfp

    @property
    def hasdist(self):
        """True if this dataset contains the particle distribution.

        Datasets from EPREM runs with `outputFlux=0`, which is the default
        setting, contain the full pitch-angle resolved particle distribution.
        This property allows users to check whether the underlying dataset
        contains the distribution (`True`) or not (`False`) before attempting to
        access the corresponding array property.

        See Also
        --------
        `~hasflux`
        """
        if self._hasdist is None:
            self._hasdist = 'Dist' in self.view.arrays
        return self._hasdist

    @property
    def hasflux(self):
        """True if this dataset contains the particle flux.

        Datasets from EPREM runs with `outputFlux=1` contain pre-computed
        particle flux in place of the pitch-angle resolved particle
        distribution. This property allows users to check whether the underlying
        dataset contains pre-computed flux (`True`) or not (`False`) before
        attempting to access the corresponding array property.

        See Also
        --------
        `~hasdist`
        """
        if self._hasflux is None:
            self._hasflux = 'flux' in self.view.arrays
        return self._hasflux

    @property
    def view(self):
        """A view into this dataset on disk."""
        if self._view is None:
            self._view = view_factory(self.source)
        return self._view

    @property
    def axes(self):
        """An interface to canonical array axes."""
        if self._axes is None:
            self._axes = axes_factory(self.source, system=self.system)
        return self._axes

    @property
    def source(self):
        """This dataset's source path."""
        return self._source

    @property
    def system(self):
        """The metric system in which to represent this dataset."""
        return self._system


@etc.autostr
class Grid:
    """An interface to the coordinate arrays in a dataset.

    This is effectively a restricted version of `~Arrays` that serves as a
    convenient namespace for the coordinate-like quantities in a dataset.
    """

    def __init__(
        self,
        source: pathlib.Path,
        system: metric.System,
    ) -> None:
        """Initialize this interface."""
        self._arrays = arrays_factory(source, system)
        self._source = None
        self._system = None
        self._r = None
        self._theta = None
        self._phi = None

    @property
    def r(self):
        """The radial coordinate."""
        return self._arrays.r

    @property
    def theta(self):
        """The polar coordinate."""
        return self._arrays.theta

    @property
    def phi(self):
        """The azimuthal coordinate."""
        return self._arrays.phi

    @property
    def source(self):
        """The dataset's source path."""
        return self._arrays.source

    @property
    def system(self):
        """The dataset's metric system."""
        return self._arrays.system


def axes_factory(
    source: paths.PathLike,
    system: typing.Optional[typing.Union[str, metric.System]]=None,
) -> Axes:
    """Create an interface to all axes in a dataset."""
    return Axes(
        source=paths.fullpath(source, strict=True),
        system=metric.system(system or 'mks'),
    )


def arrays_factory(
    source: paths.PathLike,
    system: typing.Optional[typing.Union[str, metric.System]]=None,
) -> Arrays:
    """Create an interface to all arrays in a dataset."""
    return Arrays(
        source=paths.fullpath(source, strict=True),
        system=metric.system(system or 'mks'),
    )


def grid_factory(
    source: paths.PathLike,
    system: typing.Optional[typing.Union[str, metric.System]]=None,
) -> Grid:
    """Create an interface to the coordinate arrays in a dataset."""
    return Grid(source, system)





import numbers
import typing

from .. import measured
from .. import metric
from .. import real
from .. import quantity
from ._object import (
    Object,
    ObjectT as ObjectType,
)
from ._axis import (
    Axis,
    AxisTypeError,
    AxisValueError,
    Points,
    Symbols,
    Coordinates,
)
from ._axes import (
    AxisType,
    Axes,
)
from . import _object_factories
from . import _axis_factories
from ._types import (
    ArrayType,
    QuantityType,
)
# NOTE: Importing these objects from _operations.py (instead of from their
# respective modules) has the **necessary** side-effect of running the code that
# implements all operators on physical objects.
from ._operations import (
    Scalar,
    Tensor,
    Vector,
    Array,
)


__all__ = [
    'Axis',
    'AxisTypeError',
    'AxisValueError',
    'Axes',
    'Points',
    'Symbols',
    'Coordinates',
    'Object',
    'ObjectType',
    'QuantityType',
    'ArrayType',
    'Scalar',
    'scalar',
    'Tensor',
    'tensor',
    'Vector',
]


@typing.overload
def scalar(
    x: typing.Any,
    /,
    unit: typing.Optional[metric.UnitLike]=None,
) -> Scalar:
    """Create a physical scalar from the given arguments, if possible"""

@typing.overload
def scalar(
    x: typing.Union[real.ValueType, typing.Sequence[real.ValueType]],
    /,
) -> Scalar[real.ValueType]:
    """Create a unitless physical scalar from a real number.

    Parameters
    ----------
    x : real number or equivalent length-1 sequence
        The numeric value of the new scalar, or a length-1 sequence containing
        the numeric value of the new scalar.

    Returns
    -------
    `~Scalar`
        A new physical scalar with the given value, and a unit equal to '1'.

    Raises
    ------
    `TypeError`
        The argument is a sequence with length > 1.
    """

@typing.overload
def scalar(
    x: typing.Union[real.ValueType, typing.Sequence[real.ValueType]],
    /,
    unit: metric.UnitLike,
) -> Scalar[real.ValueType]:
    """Create a physical scalar from a real number and a unit.

    Parameters
    ----------
    x : real number or equivalent length-1 sequence
        The numeric value of the new scalar, or a length-1 sequence containing
        the numeric value of the new scalar.
    unit : unit-like
        The metric unit associated with `x`.

    Returns
    -------
    `~Scalar`
        A new physical scalar with the given value and unit.

    Raises
    ------
    `TypeError`
        The argument to `x` is a sequence with length > 1.
    """

@typing.overload
def scalar(x: Scalar[real.ValueType], /) -> Scalar[real.ValueType]:
    """Create a physical scalar from a physical scalar.

    Parameters
    ----------
    x : `~Scalar`
        The scalar from which to create a new scalar. This function will make a
        copy of `x`.

    Returns
    -------
    `~Scalar`
        A new physical scalar with value and unit taken from the given scalar.

    Raises
    ------
    `ValueError`
        The caller passed an explicit unit.
    """

@typing.overload
def scalar(x: Tensor[real.ValueType], /) -> Scalar[real.ValueType]:
    """Create a physical scalar from a physical tensor.

    Parameters
    ----------
    x : `~Tensor`
        The tensor from which to create a new scalar. The tensor must be
        singular (i.e., its data must comprise a single value).

    Returns
    -------
    `~Scalar`
        A new physical scalar with value and unit taken from the given tensor.

    Raises
    ------
    `TypeError`
        The given tensor is not singular.
    `ValueError`
        The caller passed an explicit unit.
    """

def scalar(x, /, unit=None):
    return _object_factories.scalar_factory(x, unit=unit)


@typing.overload
def vector(
    x: typing.Any,
    /,
    unit: typing.Optional[metric.UnitLike]=None,
) -> Vector:
    """Create a physical vector from the given arguments, if possible"""

@typing.overload
def vector(x: typing.Sequence[real.ValueType], /) -> Vector[real.ValueType]:
    """Create a unitless physical vector from a real-valued sequence.

    Parameters
    ----------
    x : sequence of `int`, `float`, or `numbers.Real`
        The numeric value(s) of the new vector. This function will convert the
        given sequence to a `numpy.ndarray` if necessary.

    Returns
    -------
    `~Vector`
        A new physical vector with the given value(s), and a unit equal to
        '1'.

    Raises
    ------
    `TypeError`
        The input is not logically 1-dimensional.
    """

@typing.overload
def vector(
    x: typing.Sequence[real.ValueType],
    /,
    unit: metric.UnitLike,
) -> Vector[real.ValueType]:
    """Create a physical vector from a real-valued sequence and a unit.

    Parameters
    ----------
    x : sequence of `int`, `float`, or `numbers.Real`
        The numeric value(s) of the new vector. This function will convert the
        given sequence to a `numpy.ndarray` if necessary.
    unit : unit-like
        The metric unit associated with `x`.

    Returns
    -------
    `~Vector`
        A new physical vector with the given value(s) and unit.

    Raises
    ------
    `TypeError`
        The input is not logically 1-dimensional.
    """

@typing.overload
def vector(x: Vector[real.ValueType], /) -> Vector[real.ValueType]:
    """Create a physical vector from an existing physical vector.

    Parameters
    ----------
    x : `~Vector`
        The vector from which to create a new vector. This function will
        make a copy of `x`.

    Returns
    -------
    `~Vector`
        A new physical vector with value and unit taken from the given
        vector.

    Raises
    ------
    `ValueError`
        The caller passed an explicit unit.
    """

@typing.overload
def vector(x: Tensor[real.ValueType], /) -> Vector[real.ValueType]:
    """Create a physical vector from an existing physical tensor.

    Parameters
    ----------
    x : `~Tensor`
        A 1-dimensional tensor from which to create a new vector.

    Returns
    -------
    `~Vector`
        A new physical vector with value and unit taken from the given tensor.

    Raises
    ------
    `ValueError`
        The caller passed an explicit unit.
    """

@typing.overload
def vector(x: Scalar[real.ValueType], /) -> Vector[real.ValueType]:
    """Create a physical vector from an existing physical scalar.

    Parameters
    ----------
    x : `~Scalar`
        The scalar from which to create a new vector. This function will
        convert the scalar's data value into a 1-element `numpy.ndarray`.

    Returns
    -------
    `~Vector`
        A new physical vector with value and unit taken from the given scalar.

    Raises
    ------
    `ValueError`
        The caller passed an explicit unit.
    """

@typing.overload
def vector(
    x: quantity.Measurement[real.ValueType],
    /,
) -> Vector[real.ValueType]:
    """Create a physical vector from an existing measurement.

    Parameters
    ----------
    x : `~quantity.Measurement`
        The measurement from which to create a new vector. This function will
        convert the measurement's data into a `numpy.ndarray` if necessary.

    Returns
    -------
    `~Vector`
        A new physical vector with data and unit taken from the given
        measurement.

    Raises
    ------
    `ValueError`
        The caller passed an explicit unit.
    """

def vector(x, /, unit=None):
    return _object_factories.vector_factory(x, unit=unit)


@typing.overload
def tensor(
    x: typing.Any,
    /,
    unit: typing.Optional[metric.UnitLike]=None,
) -> Tensor:
    """Create a physical tensor from the given arguments, if possible"""

@typing.overload
def tensor(x: typing.Sequence[real.ValueType], /) -> Tensor[real.ValueType]:
    """Create a unitless physical tensor from a real-valued sequence.

    Parameters
    ----------
    x : sequence of `int`, `float`, or `numbers.Real`
        The numeric value(s) of the new tensor. This function will convert the
        given sequence to a `numpy.ndarray` if necessary.

    Returns
    -------
    `~Tensor`
        A new physical tensor with the given value(s), and a unit equal to '1'.

    Raises
    ------
    `TypeError`
        The input produced a rank-0 tensor.
    """

@typing.overload
def tensor(
    x: typing.Sequence[real.ValueType],
    /,
    unit: metric.UnitLike,
) -> Tensor[real.ValueType]:
    """Create a physical tensor from a real-valued sequence and a unit.

    Parameters
    ----------
    x : sequence of `int`, `float`, or `numbers.Real`
        The numeric value(s) of the new tensor. This function will convert the
        given sequence to a `numpy.ndarray` if necessary.
    unit : unit-like
        The metric unit associated with `x`.

    Returns
    -------
    `~Tensor`
        A new physical tensor with the given value(s) and unit.

    Raises
    ------
    `TypeError`
        The input produced a rank-0 tensor.
    """

@typing.overload
def tensor(x: Tensor[real.ValueType], /) -> Tensor[real.ValueType]:
    """Create a physical tensor from an existing physical tensor.

    Parameters
    ----------
    x : `~Tensor`
        The tensor from which to create a new tensor. This function will make a
        copy of `x`.

    Returns
    -------
    `~Tensor`
        A new physical tensor with value and unit taken from the given tensor.

    Raises
    ------
    `ValueError`
        The caller passed an explicit unit.
    """

@typing.overload
def tensor(x: Scalar[real.ValueType], /) -> Tensor[real.ValueType]:
    """Create a physical tensor from an existing physical scalar.

    Parameters
    ----------
    x : `~Scalar`
        The scalar from which to create a new tensor. This function will convert
        the scalar's data value into a 1-element `numpy.ndarray`.

    Returns
    -------
    `~Tensor`
        A new physical tensor with value and unit taken from the given scalar.

    Raises
    ------
    `ValueError`
        The caller passed an explicit unit.
    """

@typing.overload
def tensor(
    x: quantity.Measurement[real.ValueType],
    /,
) -> Tensor[real.ValueType]:
    """Create a physical tensor from an existing measurement.

    Parameters
    ----------
    x : `~quantity.Measurement`
        The measurement from which to create a new tensor. This function will
        convert the measurement's data into a `numpy.ndarray` if necessary.

    Returns
    -------
    `~Tensor`
        A new physical tensor with data and unit taken from the given
        measurement.

    Raises
    ------
    `ValueError`
        The caller passed an explicit unit.
    """

def tensor(x, /, unit=None):
    return _object_factories.tensor_factory(x, unit=unit)


@typing.overload
def points(
    x: typing.Union[
        typing.SupportsInt,
        typing.Sequence[typing.SupportsInt],
        Points,
    ],
    /,
) -> Points:
    """Create an integral axis.

    Parameters
    ----------
    TODO
    """

def points(x, /):
    return _axis_factories.points_factory(x)

@typing.overload
def symbols(
    x: typing.Union[str, typing.Sequence[str], Symbols],
    /,
) -> Symbols:
    """Create a symbolic axis.

    Parameters
    ----------
    TODO
    """

def symbols(x, /):
    return _axis_factories.symbols_factory(x)


@typing.overload
def coordinates(
    x: typing.Union[
        numbers.Real,
        typing.Sequence[numbers.Real],
        Coordinates,
    ],
    /,
) -> Coordinates:
    """Create a coordinate axis.

    Parameters
    ----------
    TODO
    """

def coordinates(x, /):
    return _axis_factories.coordinates_factory(x)


@typing.overload
def axes(
    mapping: typing.Mapping[str, AxisType],
    *lengths: numbers.Integral,
    shape: typing.Optional[typing.Iterable[numbers.Integral]]=None,
    dimensions: typing.Optional[typing.Iterable[str]]=None,
    axes: typing.Optional[typing.Iterable[AxisType]]=None,
    **pairs: AxisType
) -> Axes:
    """Create a axes from arguments. See Notes for precedence rules.

    Returns
    -------
    `~axes`

    Raises
    ------
    `ValueError`
        The given combination of arguments produced an empty axes.
    `TypeError`
        The caller passed `shape` or `axes` as a positional argument.

    Notes
    -----
    - A axes initialized from only a shape will contain axis pairs in which
      each name has the form `xi` and the corresponding sequence is a length-`i`
      zero-based integer array, where `i` ranges from 0 to `len(shape)-1`.
    - A axes initialized from a sequence of anonymous axes will contain
      key-value pairs with keys constructed as if initialized from a shape, each
      corresponding to an axis in the order given.
    - Explicit dimensions will replace default axis names in a axes initialized
      from either a shape or axes.
    - A axes initialized from key-value pairs or from a single mapping will use
      each key as a dimension and the corresponding value as the axis object.
    - In case of multiple initialization styles, key-value pairs take precedence
      over a mapping, which takes precedence over axes (with or without explicit
      dimensions), which take precedence over shape (with or without explicit
      dimensions).
    """

@typing.overload
def axes(*lengths: numbers.Integral) -> Axes:
    """Create a axes with the given axis lengths.

    Parameters
    ----------
    *lengths : integral numbers
        One or more integral axis lengths.

    Returns
    -------
    `~axes`
    """

@typing.overload
def axes(
    *,
    shape: typing.Iterable[numbers.Integral],
    dimensions: typing.Iterable[str],
) -> Axes:
    """Create a axes with the given shape.

    Parameters
    ----------
    shape : iterable of integral numbers
        An iterable object containing one or more integral axis lengths. Must be
        passed as a keyword argument.
    dimensions : iterable of strings
        The name of each axis. The length of `dimensions` must equal the length
        of `shape`. Must be passed as a keyword argument.

    Returns
    -------
    `~axes`

    Raises
    ------
    `TypeError`
        Caller passed dimensions without shape.
    `ValueError`
        Caller passed dimensions and shape with different lengths.
    """

@typing.overload
def axes(
    *,
    dimensions: typing.Iterable[str],
    axes: typing.Iterable[AxisType],
) -> Axes:
    """Create a axes from anonymous axes.

    Parameters
    ----------
    dimensions : iterable of strings
        The name of each axis. The length of `dimensions` must equal the length
        of `axes`. Must be passed as a keyword argument.
    axes : iterable of axis-like
        An iterable object containing one or more axis-like objects. Each
        argument may be an iterable of integers or an instance of `~Axis`. Must
        be passed as a keyword argument.

    Returns
    -------
    `~axes`

    Raises
    ------
    `TypeError`
        Caller passed dimensions without shape.
    `ValueError`
        Caller passed dimensions and shape with different lengths.
    """

@typing.overload
def axes(**pairs: AxisType) -> Axes:
    """Create a axes from individual dimension-axis pairs.

    Parameters
    ----------
    **pairs
        One or more key-value pairs. Each value may be an iterable of integers
        or an instance of `~Axis`. The corresponding key will become the string
        dimension corresponding to that axis.

    Returns
    -------
    `~axes`
    """

@typing.overload
def axes(mapping: typing.Mapping[str, AxisType]) -> Axes:
    """Create a axes from a mapping of dimension-axis pairs.

    Parameters
    ----------
    mapping
        A mapping from dimension name to axis value. Each value may be an
        iterable of integers or an instance of `~Axis`. The corresponding key
        will become the string dimension corresponding to that axis.

    Returns
    -------
    `~axes`
    """

def axes(*args, **kwargs):
    return _axes.axes_factory(*args, **kwargs)


DimensionsLike = typing.Iterable[str]

AxesLike = typing.Mapping[str, AxisType]

@typing.overload
def array(x: typing.Any, **kwargs) -> Array:
    """Create a physical array from the given arguments, if possible"""

@typing.overload
def array(
    x: typing.Sequence[real.ValueType],
    /,
    unit: typing.Optional[metric.UnitLike]=None,
    axes: typing.Optional[typing.Union[DimensionsLike, AxesLike]]=None,
) -> Array:
    """Create a physical array from a real-valued sequence.

    Parameters
    ----------
    x : array-like sequence of real numbers
        The numerical data of the new array.
    unit : unit-like
        The metric unit associated with values in `x`.
    axes : dimensions-like or axes-like
        Either an iterable of string dimensions (i.e., axis names) or a mapping
        from dimension to axis vector. The number of axes must equal the number
        of logical dimensions in `x`.

    Returns
    -------
    `~Array`
    """

@typing.overload
def array(
    x: measured.Object[real.ValueType],
    /,
    axes: typing.Optional[typing.Union[DimensionsLike, AxesLike]]=None,
) -> Array:
    """Create a physical array from a measured object.

    Parameters
    ----------
    x : `~measured.Object`
        The measured object from which to take numerical data and the associated
        metric unit. If `x` is a scalar object, this function will convert it to
        a 1-D object.
    axes : dimensions-like or axes-like
        Either an iterable of string dimensions (i.e., axis names) or a mapping
        from dimension to axis vector. The number of axes must equal the number
        of dimensions in `x`.

    Returns
    -------
    `~Array`

    Raises
    ------
    `ValueError`
        The caller passed an explicit unit.
    """

@typing.overload
def array(x: Array[real.ValueType], /) -> Array[real.ValueType]:
    """Create a physical array from an existing physical array."""

def array(*args, **kwargs):
    return _object_factories.array_factory(*args, **kwargs)





import typing

import numpy

from .. import base
from .. import measured
from ..typehelp import Self
from ._object import Object
from ._scalar import Scalar
from ._types import QuantityType as Q


class Vector(Object[base.SequenceType]):
    """A one-dimensional physical object."""

    def __init__(
        self,
        data: base.SequenceType,
        context: measured.Context[Q],
        /,
    ) -> None:
        array = numpy.asanyarray(data)
        super().__init__(array, context)
        self._size = array.size

    def __contains__(self, v) -> bool:
        """Called for v in self."""
        return v in self.data

    def __len__(self) -> int:
        """Called for len(self)."""
        return len(self.data)

    def __iter__(self):
        """Called for iter(self)."""
        return (Scalar.spawn(x, unit=self.unit) for x in self.data)

    @typing.overload
    def __getitem__(self, __i: int) -> Scalar: ...

    @typing.overload
    def __getitem__(self: Self, __i: slice) -> Self: ...

    def __getitem__(self, __i):
        """Called for self[__i]."""
        data = self.data[__i]
        try:
            result = Scalar.spawn(data, unit=self.unit)
        except TypeError:
            result = self.spawn(data, unit=self.unit)
        return result

    def __array__(self, *args, **kwargs) -> numpy.ndarray:
        """Called for conversion to a native `numpy` array."""
        return numpy.array(self.data, *args, **kwargs)

    @property
    def size(self) -> int:
        """The number of data values in this vector."""
        return self._size

    def _get_numpy_array(self):
        return numpy.array(self.data)

    _FUNCTIONS = {}
    _OPERATORS = {}





import contextlib
import numbers
import typing
import typing_extensions

import numpy

from .. import base
from .. import container
from .. import etc
from .. import measured
from .. import metric
from .. import numeric
from .. import real
from ._exceptions import (
    ParsingTypeError,
    ParsingValueError,
)


def isindexlike(x, /):
    """True if the input can index an axis."""
    if isinstance(x, (numeric.index.Value, numeric.index.Sequence)):
        return True
    for f in (numeric.index.value, numeric.index.sequence):
        with contextlib.suppress(TypeError, ValueError):
            f(x)
            return True
    return False


def ismeasurable(x, /):
    """True if the user can expect to be able to measure the input.

    A measurable object may be:
    
    - an object that satisfies the `~quantity.Measurable` protocol
    - a real number
    - an iterable of real numbers
    - an iterable of real numbers followed by a unit-like object
    - an two-element iterable whose first element is an iterable of real numbers
      and whose second element is a unit-like object
    - an iterable of any of the previous objects.

    Parameters
    ----------
    x
        The candidate measurable object.

    Returns
    -------
    bool
        True if `x` is measurable; false otherwise.

    See Also
    --------
    `~measure`
        Create a `~Measurement` from measurable input.
    """
    args = container.unwrap(x)
    if etc.isnull(x):
        return False
    if base.ismeasurable(args):
        return True
    if isinstance(args, numbers.Real):
        return True
    if not container.isseparable(args):
        return False
    if all(isinstance(arg, numbers.Real) for arg in args):
        return True
    if metric.unitlike(args[-1]):
        arg0 = args[0]
        values = arg0 if container.isiterable(arg0) else args[:-1]
        if all(isinstance(value, numbers.Real) for value in values):
            return True
    if all(isinstance(i, (list, tuple)) and ismeasurable(i) for i in args):
        return True
    return False


Ns = typing_extensions.TypeVarTuple('Ns')

@typing.overload
def parse(
    x: real.ValueType,
    /,
) -> typing.Tuple[real.ValueType, typing.Literal['1']]: ...

@typing.overload
def parse(
    x: typing.Tuple[real.ValueType, str],
    /,
) -> typing.Tuple[real.ValueType, str]: ...

@typing.overload
def parse(
    x: typing.Tuple[real.ValueType, ...],
    /,
) -> typing.Tuple[typing_extensions.Unpack[Ns], typing.Literal['1']]: ...

@typing.overload
def parse(
    x: typing.Tuple[typing_extensions.Unpack[Ns], str],
    /,
) -> typing.Tuple[typing_extensions.Unpack[Ns], str]: ...

@typing.overload
def parse(
    x: typing.Tuple[typing_extensions.Unpack[Ns], str],
    /,
    distribute: typing.Literal[False],
) -> typing.Tuple[typing_extensions.Unpack[Ns], str]: ...

@typing.overload
def parse(
    x: typing.Tuple[typing_extensions.Unpack[Ns], str],
    /,
    distribute: typing.Literal[True],
) -> typing.Tuple[typing.Tuple[numbers.Real, str], ...]: ...

def parse(x, /, distribute: bool=False):
    """Parse an implicitly measurable object.
    
    This function will extract numeric values and an associated unit from `x`,
    if possible.

    Parameters
    ----------
    x
        The object or collection of objects to measure.

    distribute : bool, default=false
        If true, distribute the parsed unit among parsed numerical values.

    Returns
    -------
    `tuple`
        - If `distribute == True`, the returned tuple will contain one or more
          `tuple`s, each containing a single numeric value and a string that
          represents the associated unit common to all values. The length of the
          result will equal the number of parsed numeric values.
        - Otherwise, the returned `tuple` will contain parsed numeric values
          followed by a single unit string. The length of the result will be one
          more than the number of original numeric values.

    Raises
    ------
    `~ParsingTypeError`
        - The input is any non-iterable object but is not a single number.

    `~ParsingValueError`
        - The input is an empty iterable object.
        - The input contains multiple units.
        - The input contains multiple objects with individual units, but the
          units differ.

    Notes
    -----
    This function will parse the following types of arguments

    - a single number
    - a `list` or `tuple` of only numbers
    - a `list` or `tuple` of numbers followed by a single unit
    - a 2-element `list` or `tuple` in which the first element is a `list` or
      `tuple` of only numbers, and the second element is a single unit
    - an iterable collection of any of the above cases, as long as the units are
      consistent

    where the notion of numbers includes strings that can be converted to
    numeric values.
    """

    # Strip redundant lists and tuples.
    unwrapped = container.unwrap(x)

    # Raise a type-based exception if input is `None`.
    if unwrapped is None:
        raise ParsingTypeError(f"Cannot measure {unwrapped!r}") from None

    # Raise a value-based exception for empty input.
    if etc.isnull(unwrapped):
        raise ParsingValueError(
            f"Cannot measure empty input: {unwrapped!r}"
        ) from None

    # Handle a single numeric value.
    if isinstance(unwrapped, numbers.Real):
        result = (unwrapped, '1')
        return (result,) if distribute else result

    # Handle a single numerical string.
    if isinstance(unwrapped, str):
        try:
            result = (float(unwrapped), '1')
        except (ValueError, TypeError) as err:
            raise ParsingTypeError(
                f"Cannot measure non-numeric string {unwrapped!r}"
            ) from err
        return (result,) if distribute else result

    # Raise a type-based exception if input is not iterable.
    try:
        iter(unwrapped)
    except TypeError as err:
        raise ParsingTypeError(
            f"Cannot measure non-iterable input: {unwrapped!r}"
        ) from err

    # Recursively parse nested parsable objects.
    if all(isinstance(arg, (list, tuple)) for arg in unwrapped):
        return _recursive_parse(unwrapped, distribute)

    # Count the number of distinct unit-like objects.
    types = [type(arg) for arg in unwrapped]
    counted = {t: types.count(t) for t in (str, metric.Unit)}

    # Check for multiple units.
    errmsg = "You may only specify one unit."
    if counted[metric.Unit] > 1:
        # If the input contains more than one instance of the `Unit` class,
        # there is nothing we can do to salvage it.
        raise ParsingValueError(errmsg) from None
    if counted[str] > 1:
        # First, check for a single numeric string.
        if isinstance(unwrapped, str):
            with contextlib.suppress(ValueError):
                return parse([float(unwrapped)])
        # Next, check for all numeric strings.
        with contextlib.suppress(ValueError):
            return parse([float(arg) for arg in unwrapped])
        # Finally, check for numeric strings with a final unit.
        try:
            values = [float(arg) for arg in unwrapped[:-1]]
        except ValueError as err:
            raise ParsingValueError(errmsg) from err
        return parse([*values, unwrapped[-1]])

    # Handle flat numerical iterables, like (1.1,) or (1.1, 2.3).
    if all(isinstance(arg, numbers.Real) for arg in unwrapped):
        return _wrap_measurable(unwrapped, '1', distribute)

    # Ensure an explicit unit-like object. Note that, at this point, `unwrapped`
    # must have one of the following forms (where any tuple may be a list):
    # - (v0, v1, ..., unit)
    # - ((v0, v1, ...), unit)
    last = unwrapped[-1]
    unitless = all(
        not isinstance(arg, (str, metric.Unit)) for arg in unwrapped
    ) or last in ['1', metric.unit('1')]
    unit = '1' if unitless else str(last)

    if isinstance(last, (str, metric.Unit)):
        unit = str(unwrapped[-1])

        # Handle flat iterables with a unit, like (1.1, 'm') or (1.1, 2.3, 'm').
        if all(isinstance(arg, numbers.Real) for arg in unwrapped[:-1]):
            return _wrap_measurable(unwrapped[:-1], unit, distribute)

        # Handle iterable values with a unit, like [(1.1, 2.3), 'm'].
        if isinstance(unwrapped[0], (list, tuple, range)):
            return _wrap_measurable(unwrapped[0], unit, distribute)

    raise ParsingTypeError(x)


def _wrap_measurable(values, unit, distribute: bool):
    """Wrap a parsed measurable and return to caller."""
    if distribute:
        return tuple(container.distribute(values, unit))
    return (*values, unit)


def _recursive_parse(unwrapped, distribute: bool):
    """Parse the measurable by calling back to `~parse`."""
    if distribute:
        parsed = [
            item
            for arg in unwrapped
            for item in parse(arg, distribute=True)
        ]
        return tuple(parsed)
    parsed = [
        parse(arg, distribute=False) for arg in unwrapped
    ]
    units = [item[-1] for item in parsed]
    if any(unit != units[0] for unit in units):
        errmsg = "Cannot combine measurements with different units."
        raise ParsingValueError(errmsg)
    values = [
        i for item in parsed for i in item[:-1]
    ]
    unit = units[0]
    return (*values, unit)


def hasdata(a):
    """True if the argument appears to contain numeric data."""
    return getdata(a) is not None # NOTE: data could be 0


def getdata(a):
    """Get the argument's numeric data, if any."""
    if isinstance(a, measured.Object):
        return a.data
    if isinstance(a, (numbers.Number, numpy.number, numpy.ndarray)):
        return a
    with contextlib.suppress(ParsingTypeError, ParsingValueError):
        parsed = parse(a, distribute=False)
        data = parsed[:-1]
        if len(data) > 1:
            return numpy.array(data)
        return data[0]





import abc
import numpy

from .. import base
from .. import exceptions
from .. import measured
from .. import numeric
from .. import real
from ._exceptions import PhysicalTypeError
from ._object import (
    Object,
)
from ._types import QuantityType as Q


class Operators(abc.ABC):
    """Abstract base class for physical scalar operators."""

    @abc.abstractmethod
    def __round__(self: Q, ndigits: int=None) -> Q:
        """Called for round(self)."""

    @abc.abstractmethod
    def __floor__(self: Q) -> Q:
        """Called for math.floor(self)."""

    @abc.abstractmethod
    def __ceil__(self: Q) -> Q:
        """Called for math.ceil(self)."""

    @abc.abstractmethod
    def __trunc__(self: Q) -> Q:
        """Called for math.trunc(self)."""


class Scalar(Object[real.ValueType], Operators):
    """A single-valued physical object."""

    def __init__(
        self,
        data: real.ValueType,
        context: measured.Context[Q],
        /,
    ) -> None:
        if not isinstance(data, base.Scalar):
            raise TypeError(
                f"Data argument to {type(self)} must implement"
                f" the {base.Scalar} protocol"
            ) from None
        super().__init__(data, context)

    @exceptions.convert(PhysicalTypeError, TypeError)
    def __int__(self) -> int:
        """Called for int(self)."""
        if operator := self._OPERATORS.get(int):
            return operator(self)
        raise TypeError(
            f"Instances of {type(self)} do not support"
            f" conversion to built-in {int} type"
        ) from None

    @exceptions.convert(PhysicalTypeError, TypeError)
    def __float__(self) -> float:
        """Called for float(self)."""
        if operator := self._OPERATORS.get(float):
            return operator(self)
        raise TypeError(
            f"Instances of {type(self)} do not support"
            f" conversion to built-in {float} type"
        ) from None

    @exceptions.convert(PhysicalTypeError, TypeError)
    def __complex__(self) -> complex:
        """Called for complex(self)."""
        if operator := self._OPERATORS.get('complex'):
            return operator(self)
        raise TypeError(
            f"Instances of {type(self)} do not support"
            f" conversion to built-in {complex} type"
        ) from None

    __round__ = numeric.implementation.unary('round')
    __floor__ = numeric.implementation.unary('floor')
    __ceil__ = numeric.implementation.unary('ceil')
    __trunc__ = numeric.implementation.unary('trunc')

    def _get_numpy_array(self):
        return numpy.array(self.data)

    _FUNCTIONS = {}
    _OPERATORS = {}





import typing

import numpy

from .. import base
from .. import container
from .. import measured
from .. import numeric
from .. import real


T = typing.TypeVar('T')


class Measurement(measured.Sequence[real.ValueType]):
    """A physical measurement.
    
    An instance of this class behaves like a numeric sequence, and additionally
    supports converting single-valued instances to built-in `int`, `float`, or
    `complex` type. It also inherits unit support from `~measured.Object`.

    See Also
    --------
    `~measured.Sequence`
        A sequence of real values with an associated metric unit.
    """

    def __int__(self) -> int:
        """Called for int(self)."""
        return self._cast_to(int)

    def __float__(self) -> float:
        """Called for float(self)."""
        return self._cast_to(float)

    def __complex__(self) -> complex:
        """Called for complex(self)."""
        return self._cast_to(complex)

    def _cast_to(self, __type: typing.Type[T]) -> T:
        """Internal logic for casting data to type `T`."""
        if self.size == 1:
            return __type(self.data[0])
        raise TypeError(
            f"Cannot convert size-{self.size} sequence to float"
        ) from None


@Measurement.register.factory
def measurement_factory(x, /, unit=None):
    """Factory function for `~Measurement`."""
    d, u = measurement_args(x, unit)
    ctx = measured.context(u, measured.convert.arraylike)
    return Measurement(d, ctx)


def measurement_args(x, unit, /):
    if isinstance(x, measured.Object):
        sequence = measured.sequence(x)
        return sequence.data, sequence.unit
    if container.isseparable(x):
        if numeric.data.hasdtype(x, (numpy.integer, numpy.floating)):
            return numpy.array(x), unit
        raise TypeError(
            f"Cannot create a measurement from {x!r}"
            " with non-numeric values."
        )
    if not base.isreal(x):
        raise TypeError(
            f"Cannot create measurement from object of type {type(x)}"
        ) from None
    return numpy.array([x]), unit


